(* avltree.sml
 _______________________________________________________________________
|									|
|			      Projet LCS				|
|			  Groupe OLC, CNRS/LAAS				|
|			 7, Av. du Colonel Roche			|
|			   31077 Toulouse Cedex				|
|_______________________________________________________________________|

AVL search trees with string keys.

*)



(* signature

structure Avltree =
  abstype 'a stree
  val nullstree : 'a stree
  exception Lookupstree : exn
  val lookupstree : 'a stree -> string -> 'a
  val updatestree : 'a stree * (string * 'a) -> 'a stree
  val streedepth : 'a stree -> int
  val balancestree : 'a stree -> 'a stree
  val list_of_stree : 'a stree -> (string * 'a) list
  val mapstree : ('a -> 'b) -> 'a stree -> 'b stree

*)



structure Avltree =
struct

abstype 'a stree = (* with strings keys, could be parameterized *)
	  null
	| stree of string * 'a * int * 'a stree * 'a stree
with
val nullstree = null;
exception Lookupstree;
fun lookupstree tr k = (* k is the key, r points to a env *)
	(* returns ref to content for key k, k assumed to exist *)
	let fun lookup null = raise Lookupstree
	      | lookup (stree (K,C,_,tr1,tr2)) =
			if k>K then lookup tr2
			if K>k then lookup tr1
			else C
	in lookup tr
	end;
local
fun H null = 0 | H (stree(_,_,h,_,_)) = h;
fun rot_left (k,i,t1,stree(k2,i2,h2,l2,g2)) =
	let val hh = 1+max(H t1,H l2)
	in stree(k2,i2,1+max(hh,H g2),stree(k,i,hh,t1,l2),g2)
	end;
fun rot_right (k,i,stree(k1,i1,h1,l1,g1),t2) =
	let val hh = 1+max(H g1,H t2)
	in stree(k1,i1,1+max(H l1,hh),l1,stree(k,i,hh,g1,t2))
	end;
fun insert k i null = 
	stree(k,i,1,null,null)
  | insert k i (t as stree(k',i',h,le,gr)) =
	if k<k' 
	then let val t1 as stree(k1,i1,h1,l1,g1) = insert k i le
	     in if h1 - H gr > 1
		then rot_right (k',i',t1,gr)
		else stree(k',i',1+max(h1,H gr),t1,gr)
	     end
	if k>k'
	then let val t2 as stree(k2,i2,h2,l2,g2) = insert k i gr
	     in if h2 - H le > 1
		then rot_left (k',i',le,t2)
		else stree(k',i',1+max(H le,h2),le,t2)
	     end
	else stree(k,i,h,le,gr);
in
fun updatestree (tr,(k,c)) = (* extends or updates an stree *)
        (* updates content of key k if exists, otherwise creates entry *)
	insert k c tr;
end;
fun list_of_stree null = [] (* entries in alpha order of keys *)
  | list_of_stree (stree (k,c,_,tr1,tr2)) =
		list_of_stree tr1 @ (k,c)::list_of_stree tr2;
fun streedepth null = 0 (* depth of deepest item in stree *)
  | streedepth (stree(_,_,_,e1,e2)) = max (1+streedepth e1,1+streedepth e2);
fun balancestree tr = tr; (* for compat with Stree structure;
			     avl trees may be considered balanced *)

fun mapstree f null = null
  | mapstree f (stree (k,c,b,tr1,tr2)) =
	stree (k,f c,b,mapstree f tr1, mapstree f tr2);

end; (* abstype *)

(* avltrees with states
*** TO BE PROVIDED ***
*)

end; (* structure *)




